IDI_APPICON ICON "app.ico"

#define TEXTFILE 256
#define IDR_CONFIG_FILE 101
#define IDR_BAR_SPECTRUM_SHADER 102
#define IDR_NORTHERN_LIGHTS_SHADER 103
#define IDR_PASS_THROUGH_SHADER 104
#define IDR_SPECTROGRAM_SHADER 105
#define IDR_WINAMP_LINE_STYLE_SPECTRUM_SHADER 106
#define IDR_EYE_OF_PHI_SHADER 107
#define IDR_ORION_CIRCLE_SHADER 108
#define IDR_ORION_CIRCLE_ROTATE_SHADER 109
#define IDR_ORION_SATURN_SUBRING_SHADER 110
#define IDR_ORION_SATURN_CORE_SHADER 111

#define IDR_SOLARIZED_DARK_THEME 501
#define IDR_TRICOLOR_THEME 502


IDR_CONFIG_FILE TEXTFILE "../example_files/config"
IDR_BAR_SPECTRUM_SHADER TEXTFILE "../output/shaders/bar_spectrum.frag"
IDR_NORTHERN_LIGHTS_SHADER TEXTFILE "../output/shaders/northern_lights.frag"
IDR_PASS_THROUGH_SHADER TEXTFILE "../output/shaders/pass_through.vert"
IDR_SPECTROGRAM_SHADER TEXTFILE "../output/shaders/spectrogram.frag"
IDR_WINAMP_LINE_STYLE_SPECTRUM_SHADER TEXTFILE "../output/shaders/winamp_line_style_spectrum.frag"
IDR_EYE_OF_PHI_SHADER TEXTFILE "../output/shaders/eye_of_phi.frag"
IDR_ORION_CIRCLE_SHADER TEXTFILE "../output/shaders/orion_circle.frag"
IDR_ORION_CIRCLE_ROTATE_SHADER TEXTFILE "../output/shaders/orion_circle_rotate.frag"
IDR_ORION_SATURN_SUBRING_SHADER TEXTFILE "../output/shaders/orion_saturn_subring.frag"
IDR_ORION_SATURN_CORE_SHADER TEXTFILE "../output/shaders/orion_saturn_core.frag"

IDR_SOLARIZED_DARK_THEME TEXTFILE "../output/themes/solarized_dark"
IDR_TRICOLOR_THEME TEXTFILE "../output/themes/tricolor"
